import numpy as np
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import CountVectorizer
from scipy.sparse import vstack, csr_matrix

from gismo.corpus import Corpus, toy_source_text
from gismo.embedding import Embedding

class Cluster:
    """
    The 'Cluster' class is used for internal representation of hierarchical cluster. It stores
    the attributes that describe a clustering structure and provides cluster addition (e.g. merge).
    Parameters
    ----------
    indice: int
        Index of the head (main element) of the cluster.
    rank: int
        The ranking order of a cluster.
    vector: csr_matrix
        The vector representation of the cluster

    Attributes
    ----------
    indice: int
        Index of the head (main element) of the cluster.
    rank: int
        The ranking order of a cluster.
    vector: csr_matrix
        The vector representation of the cluster.
    intersection_vector: csr_matrix (deprecated)
        The vector representation of the common points of a cluster.
    members: list of int
        The indices of the cluster elements.
    focus: float in range [0.0, 1.0]
        The consistency of the cluster (higher focus means that elements are more similar).
    children: list of Cluster
        The subclusters.

    Examples
    ---------
    >>> c1 = Cluster(indice=0, rank=1, vector=csr_matrix([1.0, 0.0, 1.0]))
    >>> c2 = Cluster(indice=5, rank=0, vector=csr_matrix([1.0, 1.0, 0.0]))
    >>> c3 = c1+c2
    >>> c3.members
    [0, 5]
    >>> c3.indice
    5
    >>> c3.vector.toarray()
    array([[2., 1., 1.]])
    >>> c3.intersection_vector.toarray()
    array([[1., 0., 0.]])
    >>> c1 == sum([c1])
    True
    """

    def __init__(self, indice=None, rank=None, vector=None):
        self.indice = indice
        self.rank = rank
        self.members = [indice] if indice is not None else []
        self.focus = 1.0
        self.vector = vector
        self.intersection_vector = vector
        self.children = []

    def __add__(self, cluster):
        result = Cluster()
        if self.rank < cluster.rank:
            result.indice = self.indice
            result.rank = self.rank
        else:
            result.indice = cluster.indice
            result.rank = cluster.rank
        result.members = self.members + cluster.members
        result.focus = min(self.focus, cluster.focus)  # Don't forget external focus
        result.vector = self.vector + cluster.vector
        result.intersection_vector = self.intersection_vector.multiply(cluster.intersection_vector)
        result.children = []  # Better update sons list outside class definition
        return result

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)


def merge_clusters(cluster_list, focus=1.0):
    if len(cluster_list) < 2:
        return cluster_list[0]
    result = sum(cluster_list)
    result.focus = min(result.focus, focus)
    result.children = sorted(cluster_list, key = lambda c: c.rank)
    return result


def subspace_partition(subspace, resolution=.9):
    n, _ = subspace.shape
    similarity_matrix = cosine_similarity(subspace, subspace) - 2 * np.identity(n)
    similarity = np.max(similarity_matrix, axis=0)
    closest = np.argmax(similarity_matrix, axis=0)
    local_similarity = [similarity[closest[i]] for i in range(n)]
    partition = [{i} for i in range(n)]
    heads = np.arange(n)
    cluster_similarity = np.ones(n)
    for i in range(n):
        for j in range(n):
            if i == j:
                continue
            sij = similarity_matrix[i, j]
            master = min(heads[i], heads[j])
            slave = max(heads[i], heads[j])
            if sij >= resolution * local_similarity[i]:
                cluster_similarity[master] = min(cluster_similarity[master], sij)
                if master != slave:
                    for k in partition[slave]:
                        heads[k] = master
                    partition[master] |= partition[slave]
                    partition[slave] = set()
    return [(c, d) for c, d in zip(partition, cluster_similarity) if c]


def rec_clusterize(cluster_list, resolution):
    if len(cluster_list) == 1:
        return cluster_list[0]
    else:
        partition = subspace_partition(vstack([c.vector for c in cluster_list]), resolution)
        return rec_clusterize([merge_clusters([cluster_list[i] for i in p[0]], p[1]) for p in partition],
                              resolution)


def subspace_clusterize(subspace, resolution=.9, indices=None):
    '''
    Converts a subspace (matrix seen as a list of vectors) to a Cluster object (hierarchical clustering).

    Parameters
    ----------
    subspace: ndarray, csr_matrix
        A ``k x m`` matrix seen as a list of ``k`` ``m``-dimensional vectors sorted by importance order.
    resolution: float in range [0.0, 1.0]
        Sets the lazyness of aggregation. A 'resolution' set to 0.0 yields a one-step clustering
        (*star* structure), while a 'resolution ' set to 1.0 yields, up to tie similarities, a binary tree
        (*dendrogram*).
    indices: list, optional
        Indicates the index for each element of the subspace. Used when 'subspace'
        is extracted from a larger space (e.g. X or Y). If not set, indices are set to ``range(k)``.

    Returns
    -------
    Cluster
        A cluster whose leaves are the `k` vectors from 'subspace'.

    Example
    _________
    >>> corpus = Corpus(toy_source_text)
    >>> vectorizer = CountVectorizer(dtype=float)
    >>> embedding = Embedding(vectorizer=vectorizer)
    >>> embedding.fit_transform(corpus)
    >>> subspace = embedding.x[1:, :]
    >>> cluster = subspace_clusterize(subspace)
    >>> len(cluster.children)
    2
    >>> cluster = subspace_clusterize(subspace, resolution=.02)
    >>> len(cluster.children)
    4
    '''
    if indices is None:
        n, _ = subspace.shape
        indices = range(n)
    return rec_clusterize([Cluster(indice=r, rank=i, vector=subspace[i, :]) for i, r in enumerate(indices)],
                          resolution)



#!/usr/bin/env python
# coding: utf-8
#
# GISMO: a Generic Information Search with a Mind of its Own

import numpy as np

from scipy.sparse import csr_matrix
from numba import njit
from sklearn.feature_extraction.text import CountVectorizer

from gismo.common import MixInIO, toy_source_text
from gismo.corpus import Corpus

# 1-norm for diffusion (input is X or Y indptr and data, inplace modification)
@njit
def l1_normalize(indptr, data):
    """
    Normalize inplace the embedding (X or Y)
    :param indptr: the pointers of the vector
    :param data:  the data of the vector
    :return: the l1 norm. Data is normalized in place
    """
    n = len(indptr) - 1
    l = np.zeros(n)
    for i in range(n):
        l[i] = np.sum(data[indptr[i]:indptr[i + 1]])
        if l[i] > 0:
            data[indptr[i]:indptr[i + 1]] /= l[i]
    return l


# Note: the use of external embedding breaks a symmetry between X and Y. IDF needs to be stored if one wants to switch.

# ITF transformation
@njit
def itf_fit_transform(indptr, data, m):
    """
    Apply ITF transformation on embedding X.
    :param indptr: pointers of X
    :param data: data of X
    :param m: Number of features
    :return: None (inplace)
    """
    n = len(indptr) - 1
    log_m = np.log(1 + m)
    for i in range(n):
        data[indptr[i]:indptr[i + 1]] *= log_m - np.log(1 + indptr[i + 1] - indptr[i])


# IDF computation
@njit
def idf_fit(indptr, n):
    """
    Computes idf vector based on embedding Y
    :param indptr: pointers of Y
    :param n: Number of documents
    :return: IDF vector of size m
    """
    m = len(indptr) - 1
    idf_vector = np.log(1 + n) * np.ones(m)
    for i in range(m):
        idf_vector[i] -= np.log(1 + indptr[i + 1] - indptr[i])
    return idf_vector


# IDF transformation
@njit
def idf_transform(indptr, data, idf_vector):
    """
    Applies IDF on embedding Y
    :param indptr:
    :param data:
    :param idf_vector:
    :return: None (inplace)
    """
    m = len(indptr) - 1
    for i in range(m):
        data[indptr[i]:indptr[i + 1]] *= idf_vector[i]


@njit
def query_shape(indices, data, idf):
    # log normalization
    data[:] = 1 + np.log(data)
    # IdF
    for i, indice in enumerate(indices):
        data[i] *= idf[indice]
    # Normalization
    norm = np.sum(data)
    if norm > 0:
        data[:] /= norm
    return norm

def auto_vect(corpus):
    """
    Creates a default vectorizer according to corpus size.
    Filter words based on their word frequency.
    Keep only word having a frequency in [min_df, max_df].
    You may pass an int (#occurences) or [0, 1] float (frequency).
    :param corpus:
    :return: a CountVectorizer object
    """
    n = len(corpus)
    (min_df, max_df) = (3, .15) if n > 100 else (1, 1.0)
    return CountVectorizer(
        min_df=min_df,
        max_df=max_df,
        ngram_range=[1, 1],
        stop_words="english",
        dtype=float
    )


class Embedding(MixInIO):
    """
    Use vectorizer to compute forward and backward embeddings in a sklearn fashion

    Parameters
    ----------
    vectorizer: CountVectorizer, optional
                Custom vectorizer to override default behavior (recommended).
                Having a vectorizer adapted to the corpus is good practice
    filename: str, optional
                If set, will load embedder from file
    """
    def __init__(
            self,
            vectorizer: CountVectorizer = None,
            filename=None,
            path='.'
    ):
        if filename is not None:
            self.load(filename=filename, path=path)
        else:
            self.vect = vectorizer
#            self.document_to_str = document_to_str

            self.n = 0  # Number of documents
            self.m = 0  # Number of features
            self.x = None  # TF-IDTF X embedding of documents into features, normalized
            self.x_norm = None  # memory of X norm for hierarchical merge
            self.y = None  # Y embedding of features into documents
            self.y_norm = None  # memory of Y norm for hierarchical merge
            self.idf = None  # idf vector
            self.features = None # vocabulary list
            self._result_found = True # keep track of projection successes


    def fit_transform(self, corpus: Corpus):
        """
        Ingest a corpus of documents.
        If not yet set, a default vectorizer is created.
        Maps each word with an index.
        Compute for each word its IDF weights (fit).
        For each document, deduce its TF-IDF weights (coordinates in the embedding) (transform) (self.x).
        The transposed matrix is explicitely built because we use sparse matrices (self.y).

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit_transform(corpus)
        >>> embedding.x  # doctest: +NORMALIZE_WHITESPACE
        <5x21 sparse matrix of type '<class 'numpy.float64'>'
    	    with 25 stored elements in Compressed Sparse Row format>
    	>>> embedding.features[:8]
    	['blade', 'chinese', 'comparing', 'demon', 'folklore', 'gizmo', 'gremlins', 'inside']
        """
        if self.vect is None:
            self.vect = auto_vect(corpus)

        # THE FIT PART
        # Start with a simple CountVectorizer X
        x = self.vect.fit_transform(corpus.iterate_text())
        # Release stop_words_ from vectorizer
        self.vect.stop_words_ = None
        # Populate vocabulary
        self.features = self.vect.get_feature_names()
        # Extract number of documents and features
        (self.n, self.m) = x.shape
        # PART OF TRANSFORM, MUTUALIZED: Apply sublinear smoothing
        x.data = 1 + np.log(x.data)
        # PART OF TRANSFORM, MUTUALIZED: Apply ITF transformation
        itf_fit_transform(indptr=x.indptr, data=x.data, m=self.m)
        # Compute transposed CountVectorizer Y
        self.y = x.tocsc()
        # Compute IDF
        self.idf = idf_fit(self.y.indptr, self.n)

        # THE TRANSFORM PART
        idf_transform(indptr=self.y.indptr, data=self.y.data, idf_vector=self.idf)
        # back to x
        self.x = self.y.tocsr(copy=True)
        # Transpose y
        self.y = self.y.T
        # Normalize
        self.x_norm = l1_normalize(indptr=self.x.indptr, data=self.x.data)
        self.y_norm = l1_normalize(indptr=self.y.indptr, data=self.y.data)

    def fit(self, corpus: Corpus):
        """
        Use a corpus to shape the vectorizer

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit(corpus)
        >>> len(embedding.idf)
        21
    	>>> embedding.features[:8]
    	['blade', 'chinese', 'comparing', 'demon', 'folklore', 'gizmo', 'gremlins', 'inside']
        """
        assert corpus
        if self.vect is None:
            self.vect = auto_vect(corpus)

        # THE FIT PART
        # Start with a simple CountVectorizer X
        x = self.vect.fit_transform(corpus.iterate_text())
        # Release stop_words_ from vectorizer
        self.vect.stop_words_ = None
        # Populate vocabulary
        self.features = self.vect.get_feature_names()
        # Extract number of documents (required for idf) and features (required in fit)
        (self.n, self.m) = x.shape
        # Compute transposed CountVectorizer Y
        self.y = x.tocsc()
        # Compute IDF
        self.idf = idf_fit(self.y.indptr, self.n)

    def fit_ext(self, embedder):
        """
        Use fit from existing embedder. Useful to launch mini-gismos at sentence level.

        Parameters
        ----------
        embedder: Embedding
                  the (external) embedding to copy

        Examples
        --------
        >>> corpus=Corpus(toy_source_text)
        >>> other_embedding = Embedding()
        >>> other_embedding.fit(corpus)
        >>> embedding = Embedding()
        >>> embedding.fit_ext(other_embedding)
        >>> len(embedding.idf)
        21
    	>>> embedding.features[:8]
    	['blade', 'chinese', 'comparing', 'demon', 'folklore', 'gizmo', 'gremlins', 'inside']
        """
        self.m = embedder.m
        self.vect = embedder.vect
        self.idf = embedder.idf
        self.features = self.vect.get_feature_names()

    def transform(self, corpus: Corpus):
        """
        Ingest a corpus of documents, assuming the fit has already been done.

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit_transform(corpus)
        >>> [embedding.features[i] for i in embedding.x.indices[:8]]
        ['gizmo', 'mogwaï', 'blade', 'sentence', 'sentence', 'shadoks', 'comparing', 'gizmo']
        >>> small_corpus = Corpus(["I only talk about Yoda", "Gizmo forever!"])
        >>> embedding.transform(small_corpus)
        >>> [embedding.features[i] for i in embedding.x.indices]
        ['yoda', 'gizmo']
        """
        # The fit part
        assert corpus
        if self.vect is None:
            self.vect = auto_vect(corpus)

        # THE FIT PART
        # Start with a simple CountVectorizer X
        x = self.vect.transform(corpus.iterate_text())
        # Release stop_words_ from vectorizer
        self.vect.stop_words_ = None
        # Extract number of documents and features
        (self.n, _) = x.shape
        # PART OF TRANSFORM, MUTUALIZED: Apply sublinear smoothing
        x.data = 1 + np.log(x.data)
        # PART OF TRANSFORM, MUTUALIZED: Apply ITF transformation
        itf_fit_transform(indptr=x.indptr, data=x.data, m=self.m)
        # Compute transposed CountVectorizer Y
        self.y = x.tocsc()

        # THE TRANSFORM PART
        idf_transform(indptr=self.y.indptr, data=self.y.data, idf_vector=self.idf)
        # back to x
        self.x = self.y.tocsr(copy=True)
        # Transpose y
        self.y = self.y.T
        # Normalize
        self.x_norm = l1_normalize(indptr=self.x.indptr, data=self.x.data)
        self.y_norm = l1_normalize(indptr=self.y.indptr, data=self.y.data)


    def query_projection(self, query, return_success=False):
        """
        Project a query in the feature space

        Parameters
        ----------
        query: str
               Text to project to the feature space

        Example
        -------
        >>> corpus=Corpus(toy_source_text)
        >>> embedding = Embedding()
        >>> embedding.fit_transform(corpus)
        >>> z = embedding.query_projection("Gizmo is not Yoda but he rocks!")
        >>> for i in range(len(z.data)):
        ...    print(f"{embedding.features[z.indices[i]]}: {z.data[i]}")
        gizmo: 0.3868528072345416
        yoda: 0.6131471927654585
        >>> embedding._result_found
        True
        >>> z = embedding.query_projection("That content does not intersect toy corpus")
        >>> embedding._result_found
        False
        """
        z = self.vect.transform([query])
        norm = query_shape(indices=z.indices, data=z.data, idf=self.idf)
        if norm == 0:
            z = csr_matrix(self.idf) /  np.sum(self.idf)
            self._result_found = False
        else:
            self._result_found = True
        if return_success:
            return z, self._result_found
        else:
            return z



from numba import njit
import numpy as np


# diffusion: starting point is on the feature (words) space
# provides ranking on features (X) and documents (Y)
@njit
def jit_diffusion(x_pointers, x_indices, x_data,
                  y_pointers, y_indices, y_data,
                  z_indices, z_data,
                  x_relevance, y_relevance,
                  alpha, n_iter, offset,
                  x_fluid, y_fluid):
    n = len(x_pointers) - 1
    m = len(y_pointers) - 1

    # Reset fluids
    x_fluid[:] = 0
    y_fluid[:] = 0
    for ind, data in zip(z_indices, z_data):
        y_relevance[ind] -= data * offset  # First round penalty
        y_fluid[ind] = data

    # Core diffusion
    for turn in range(n_iter):
        for j in range(m):
            f = y_fluid[j]
            y_fluid[j] = 0.0
            if f > 0:
                y_relevance[j] += f
                x_fluid[y_indices[y_pointers[j]:y_pointers[j + 1]]] += f * alpha * y_data[y_pointers[j]:y_pointers[j + 1]]
        for i in range(n):
            f = x_fluid[i]
            x_fluid[i] = 0.0
            if f > 0:
                x_relevance[i] += f
                y_fluid[x_indices[x_pointers[i]:x_pointers[i + 1]]] += f * alpha * x_data[x_pointers[i]:x_pointers[i + 1]]

    # Don't waste the last drop of fluid, it's free!
    for i in range(m):
        y_relevance[i] += y_fluid[i]


class DIteration:
    def __init__(self, n, m, alpha=.5, n_iter=4, offset=1.0, memory=0):
        self.x_relevance = np.zeros(n)
        self.y_relevance = np.zeros(m)
        self.x_order = None
        self.y_order = None
        self.alpha = alpha
        self.n_iter = n_iter
        self.offset = offset
        self.memory = memory
        self._x_fluid = np.zeros(n)
        self._y_fluid = np.zeros(m)

    def __call__(self, x, y, z):
        self.x_relevance[:] *= self.memory
        self.y_relevance[:] *= self.memory
        jit_diffusion(x_pointers=x.indptr, x_indices=x.indices, x_data=x.data,
                      y_pointers=y.indptr, y_indices=y.indices, y_data=y.data,
                      z_indices=z.indices, z_data=z.data,
                      x_relevance=self.x_relevance, y_relevance=self.y_relevance,
                      alpha=self.alpha, n_iter=self.n_iter, offset=self.offset,
                      x_fluid=self._x_fluid, y_fluid=self._y_fluid)
        self.x_order = np.argsort(-self.x_relevance)
        self.y_order = np.argsort(-self.y_relevance)


#!/usr/bin/env python
# coding: utf-8
#
# GISMO: a Generic Information Search with a Mind of its Own

# DIteration parameters
ALPHA = .25  # diffusion attenuation
N_ITER = 4  # Number of round-trip diffusions
D_MAX = 10  # Maximum number of documents
K_MAX = 5  # Maximum number of keywords
S_MAX = 4  # Maximum number of sentences
F_MAX = 6  # Maximum number of features

# Hierarchical clustering parameters
STRETCH = 2  # Stretch factor that determines the coverage/relevance trade-off
C_MAX = 3  # Maximum depth of concepts
P_MAX = 4  # Maximum depth of document partitions

import gzip
import errno
import os
import dill as pickle

from pathlib import Path


class MixInIO:
    """
    Provide basic save/load capacities to other classes.
    """
    def save(self, filename: str, path='.', erase=False, compress=False):
        """
        Save instance to file.

        Parameters
        ----------
        filename: str
            The stem of the filename.
        path: str or Path
            The location path.
        erase:
            Should existing file be erased if it exists?
        compress:
            Should gzip compression be used?

        Returns
        -------
        None        >>> class MyValue(MixInIO):
        ...     def __init__(self, v=0):
        ...        self.value = v


        Examples
        ----------
        >>> import tempfile
        >>> v1 = ToyClass(42)
        >>> v2 = ToyClass()
        >>> v2.value
        0
        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.save(filename='myfile', compress=True, path=tmpdirname)
        ...     dir_content = [f.name for f in Path(tmpdirname).glob('*')]
        ...     v2.load(filename='myfile', path=Path(tmpdirname))
        ...     v1.save(filename='myfile', compress=True, path=tmpdirname) # doctest.ELLIPSIS
        File ...myfile.pkl.gz already exists!
        >>> dir_content
        ['myfile.pkl.gz']
        >>> v2.value
        42

        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.save(filename='myfile', path=tmpdirname)
        ...     v1.save(filename='myfile', path=tmpdirname) # doctest.ELLIPSIS
        File ...myfile.pkl already exists!

        >>> v1.value = 51
        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...     v1.save(filename='myfile', path=tmpdirname)
        ...     v1.save(filename='myfile', path=tmpdirname, erase=True)
        ...     v2.load(filename='myfile', path=tmpdirname)
        ...     dir_content = [f.name for f in Path(tmpdirname).glob('*')]
        >>> dir_content
        ['myfile.pkl']
        >>> v2.value
        51

        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...    v2.load(filename='thisfilenamedoesnotexist') # doctest.ELLIPSIS
        Traceback (most recent call last):
         ...
        FileNotFoundError: [Errno 2] No such file or directory: ...
        """
        if isinstance(path, str):
            path = Path(path)
        destination = path / Path(filename).stem
        if compress:
            destination = destination.with_suffix(".pkl.gz")
            if destination.exists() and not erase:
                print(f"File {destination} already exists!")
            else:
                with gzip.open(destination, "wb") as f:
                    pickle.dump(self, f)
        else:
            destination = destination.with_suffix(".pkl")
            if destination.exists() and not erase:
                print(f"File {destination} already exists!")
            else:
                with open(destination, "wb") as f:
                    pickle.dump(self, f)

    def load(self, filename: str, path='.'):
        """
        Load instance from file.

        Parameters
        ----------
        filename: str
            The stem of the filename.
        path: str or Path
            The location path.

        Returns
        -------
        None
        """
        if isinstance(path, str):
            path = Path(path)
        dest = path / Path(filename).with_suffix(".pkl")
        if dest.exists():
            with open(dest, 'rb') as f:
                self.__dict__.update(pickle.load(f).__dict__)
        else:
            dest = dest.with_suffix('.pkl.gz')
            if dest.exists():
                with gzip.open(dest) as f:
                    self.__dict__.update(pickle.load(f).__dict__)
            else:
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), dest)


class ToyClass(MixInIO):
    def __init__(self, v=0):
        """
        A minimal class to demonstrate the use of the MixInIO MixIN.

        Parameters
        ----------
        v: object
            A value to store.
        """
        self.value = v


toy_source_text = ['Gizmo is a Mogwaï.',
                   'This is a sentence about Blade.',
                   'This is another sentence about Shadoks.',
                   'This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda.',
                   'In chinese folklore, a Mogwaï is a demon.']

toy_source_dict = [{'title': 'First Document', 'content': 'Gizmo is a Mogwaï.'},
                   {'title': 'Second Document', 'content': 'This is a sentence about Blade.'},
                   {'title': 'Third Document', 'content': 'This is another sentence about Shadoks.'},
                   {'title': 'Fourth Document',
                    'content': 'This very long sentence, with a lot of stuff about Star Wars inside, '
                               'makes at some point a side reference to the Gremlins movie by '
                               'comparing Gizmo and Yoda.'},
                   {'title': 'Fifth Document', 'content': 'In chinese folklore, a Mogwaï is a demon.'}]



from sklearn.metrics.pairwise import cosine_similarity


def post_document(gismo, i):
    """
    Document indice to document entry

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    i : int
        document indice

    Returns
    -------
    Object
        Document i from corpus
    """
    return gismo.corpus[i]


def post_document_content(gismo, i, max_size=None):
    """
    Document indice to document content.

    Assumes that document has a 'content' key.

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    i : int
        document indice
    max_size: int, optional
        Maximum number of chars to return

    Returns
    -------
    str
        Content of document i from corpus
    """
    return gismo.corpus[i]['content'][:max_size]


def post_feature(gismo, i):
    """
    Feature indice to feature name

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    i : int
        feature indice

    Returns
    -------
    str
        Feature i from embedding
    """
    return gismo.embedding.features[i]


# Clusters
def post_document_cluster(gismo, cluster):
    """
    Convert cluster of documents into basic json

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    cluster: Cluster
        Cluster of documents

    Returns
    -------
    dict
        dictionary with keys 'document', 'focus', and recursive 'children'
    """
    return {'document': gismo.corpus[cluster.indice],
            'focus': cluster.focus,
            'children': [post_document_cluster(gismo, c) for c in cluster.children]}


def print_document_cluster(gismo, cluster, depth=""):
    """
    Print an ASCII view of a document cluster with metrics (focus, relevance, similarity)

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    cluster: Cluster
        Cluster of documents
    depth: str, optional
        Current depth string used in recursion
    """
    sim = cosine_similarity(cluster.vector, gismo.diteration.y_relevance.reshape(1, -1))[0][0]
    if len(cluster.children) == 0:
        txt = gismo.corpus.to_text(gismo.corpus[cluster.indice])
        print(f"{depth} {txt} "
              f"(R: {gismo.diteration.x_relevance[cluster.indice]:.2f}; "
              f"S: {sim:.2f})")
    else:
        print(f"{depth} F: {cluster.focus:.2f}. "
              f"R: {sum(gismo.diteration.x_relevance[cluster.members]):.2f}. "
              f"S: {sim:.2f}.")
    for c in cluster.children:
        print_document_cluster(gismo, c, depth=depth + '-')


def post_feature_cluster(gismo, cluster):
    """
    Convert feature cluster into basic json

    Parameters
    ----------
    gismo: Gismo
        Gismo instance
    cluster: Cluster
        Cluster of features

    Returns
    -------
    dict
        dictionary with keys 'feature', 'focus', and recursive 'children'
    """
    return {'feature': gismo.embedding.features[cluster.indice],
            'focus': cluster.focus,
            'children': [post_feature_cluster(gismo, c) for c in cluster.children]}


def print_feature_cluster(gismo, cluster, depth=""):
    """
        Print an ASCII view of a feature cluster with metrics (focus, relevance, similarity)

        Parameters
        ----------
        gismo: Gismo
            Gismo instance
        cluster: Cluster
            Cluster of features
        depth: str, optional
            Current depth string used in recursion
        """
    sim = cosine_similarity(cluster.vector, gismo.diteration.x_relevance.reshape(1, -1))[0][0]
    if len(cluster.children) == 0:
        print(f"{depth} {gismo.embedding.features[cluster.indice]} "
              f"(R: {gismo.diteration.y_relevance[cluster.indice]:.2f}; "
              f"S: {sim:.2f})")
    else:
        print(f"{depth} F: {cluster.focus:.2f}. "
              f"R: {sum(gismo.diteration.y_relevance[cluster.members]):.2f}. "
              f"S: {sim:.2f}.")
    for c in cluster.children:
        print_feature_cluster(gismo, c, depth=depth + '-')




#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['scipy', 'scikit-learn', 'dill', 'numpy', 'numba']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Fabien Mathieu",
    author_email='fabien.mathieu@normalesup.org',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="GISMO is a NLP tool to rank and organize a corpus of documents according to a query.",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='gismo',
    name='gismo',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/balouf/gismo',
    version='0.2.1',
    zip_safe=False,
)



#!/usr/bin/env python
# coding: utf-8
#
# GISMO: a Generic Information Search with a Mind of its Own


import numpy as np
from itertools import chain

from gismo.common import MixInIO, toy_source_text, toy_source_dict


class Corpus(MixInIO):
    """
    Corpus class, to feed to Embedding

    Parameters
    ----------
    source: list
        The list of items that make the corpus
    to_text: function
        The function that transforms an item into text
    filename: str, optional
        Load corpus from filename
    path: str or Path, optional
        Directory where the corpus is to be loaded from.

    Examples
    --------
        >>> corpus = Corpus(toy_source_text, to_text=lambda x: f"{x[:15]}...")
        >>> for c in corpus.iterate():
        ...    print(c)
        Gizmo is a Mogwaï.
        This is a sentence about Blade.
        This is another sentence about Shadoks.
        This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda.
        In chinese folklore, a Mogwaï is a demon.

        >>> for c in corpus.iterate_text():
        ...    print(c)
        Gizmo is a Mogw...
        This is a sente...
        This is another...
        This very long ...
        In chinese folk...

        A corpus object can be saved/loaded with save and load methods. The load method can be called directly
        from the constructor by providing a filename.

        >>> import tempfile
        >>> corpus1 = Corpus(toy_source_text)
        >>> with tempfile.TemporaryDirectory() as tmpdirname:
        ...    corpus1.save(filename="myfile", path=tmpdirname)
        ...    corpus2 = Corpus(filename="myfile", path=tmpdirname)
        >>> corpus2[0]
        'Gizmo is a Mogwaï.'
    """
    def __init__(self, source=None, to_text=None, filename=None, path='.'):
        if filename is not None:
            self.load(filename=filename, path=path)
        else:
            self.source = source
            self.i = 0
            self.n = 0 if source is None else len(source)
            if to_text is None:
                self.to_text = lambda x: x
            else:
                self.to_text = to_text

    def iterate_text(self, to_text=None):
        if to_text is None:
            to_text = self.to_text
        return (to_text(entry) for entry in self)

    def iterate(self):
        return (entry for entry in self.source)

    def __getitem__(self, i):
        return self.source[i]

    def __len__(self):
        return self.n

    def merge_new_source(self, new_source, doc2key=None):
        """
        Incorporate new entries from a source

        Parameters
        ----------
        new_source: list
                    source of the same type (same to_text mostly) that the current source
        doc2key: function
                 callback  that provides unique hashable Id for documents

        Examples
        --------
        >>> corpus = Corpus(toy_source_dict.copy(), to_text=lambda x: x['content'][:14])
        >>> len(corpus)
        5
        >>> new_corpus = [{"title": "Another document", "content": "I don't know what to say!"},
        ...     {'title': 'Fifth Document', 'content': 'In chinese folklore, a Mogwaï is a demon.'}]
        >>> corpus.merge_new_source(new_corpus, doc2key=lambda e: e['title'])
        >>> len(corpus)
        6
        >>> for c in corpus.iterate_text():
        ...    print(c)
        Gizmo is a Mog
        This is a sent
        This is anothe
        This very long
        In chinese fol
        I don't know w
        """
        if doc2key is None:
            print("Incremental corpus requires to provide a doc2key function")
            return self
        if self.source is None:
            self.source = []
        new_keys = {doc2key(d) for d in new_source} - {doc2key(d) for d in self.source}
        self.source += [d for d in new_source if doc2key(d) in new_keys]
        self.n = len(self.source)


class CorpusList(MixInIO):
    """
    To concatenate a list of corpi with distinct shapes and to_text

    Example
    -------
    >>> multi_corp = CorpusList([Corpus(toy_source_text, lambda x: x[:15]+"..."),
    ...                          Corpus(toy_source_dict, lambda e: e['title'])])
    >>> len(multi_corp)
    10
    >>> multi_corp[7]
    {'title': 'Third Document', 'content': 'This is another sentence about Shadoks.'}
    >>> for c in multi_corp.iterate_text():
    ...    print(c)
    Gizmo is a Mogw...
    This is a sente...
    This is another...
    This very long ...
    In chinese folk...
    First Document
    Second Document
    Third Document
    Fourth Document
    Fifth Document
    """

    def __init__(self, corpus_list=None, filename=None, path='.'):
        if filename is not None:
            self.load(filename=filename, path=path)
        else:
            if corpus_list is None or len(corpus_list) == 0:
                print("Please provide a non-empty list of corpi!")
            else:
                self.corpus_list = corpus_list
                self.cum_n = np.cumsum([len(corpus) for corpus in self.corpus_list])
                self.n = self.cum_n[-1]

    def iterate(self):
        return chain.from_iterable([corpus.iterate() for corpus in self.corpus_list])

    def iterate_text(self):
        return chain.from_iterable([corpus.iterate_text() for corpus in self.corpus_list])

    def __getitem__(self, i):
        corpus_indice = np.searchsorted(self.cum_n, i, side='right')
        local_i = i if corpus_indice == 0 else (i - self.cum_n[corpus_indice - 1])
        return self.corpus_list[corpus_indice][local_i]

    def __len__(self):
        return self.n



"""Main module."""

from scipy.sparse import vstack, csr_matrix
from sklearn.feature_extraction.text import CountVectorizer
from functools import partial

from gismo.common import MixInIO, toy_source_dict
from gismo.corpus import Corpus
from gismo.embedding import Embedding
from gismo.diteration import DIteration
from gismo.clustering import subspace_clusterize
from gismo.post_processing import post_document, post_document_content, post_document_cluster, \
    post_feature, post_feature_cluster, print_document_cluster, print_feature_cluster


class Gismo(MixInIO):
    """
    Example
    -------
    >>> corpus = Corpus(toy_source_dict, lambda x: x['content'])
    >>> vectorizer = CountVectorizer(dtype=float)
    >>> embedding = Embedding(vectorizer=vectorizer)
    >>> embedding.fit_transform(corpus)
    >>> gismo = Gismo(corpus, embedding)
    >>> gismo.embedding.m
    36
    >>> gismo.post_document = partial(post_document_content, max_size=42)
    >>> gismo.diteration.alpha = .7
    >>> gismo.rank("Gizmo")
    >>> gismo.get_ranked_documents(3)
    ['Gizmo is a Mogwaï.', 'This very long sentence, with a lot of stu', 'In chinese folklore, a Mogwaï is a demon.']
    >>> gismo.diteration.alpha = .8
    >>> gismo.rank("Gizmo")
    >>> gismo.get_ranked_documents(3)
    ['Gizmo is a Mogwaï.', 'In chinese folklore, a Mogwaï is a demon.', 'This very long sentence, with a lot of stu']
    >>> gismo.get_ranked_features(5)
    ['mogwaï', 'gizmo', 'is', 'in', 'chinese']
    >>> gismo.post_document_cluster = print_document_cluster
    >>> gismo.get_clustered_ranked_documents() # doctest: +NORMALIZE_WHITESPACE
     F: 0.05. R: 1.85. S: 0.99.
    - F: 0.68. R: 1.77. S: 0.98.
    -- Gizmo is a Mogwaï. (R: 1.23; S: 0.98)
    -- In chinese folklore, a Mogwaï is a demon. (R: 0.27; S: 0.72)
    -- This very long sentence, with a lot of stuff about Star Wars inside, makes at some point a side reference to the Gremlins movie by comparing Gizmo and Yoda. (R: 0.26; S: 0.67)
    - F: 0.70. R: 0.08. S: 0.19.
    -- This is a sentence about Blade. (R: 0.04; S: 0.17)
    -- This is another sentence about Shadoks. (R: 0.04; S: 0.17)
    >>> gismo.post_feature_cluster = print_feature_cluster
    >>> gismo.get_clustered_ranked_features() # doctest: +NORMALIZE_WHITESPACE
     F: 0.01. R: 1.25. S: 0.94.
    - F: 0.08. R: 1.22. S: 0.93.
    -- F: 0.99. R: 1.03. S: 0.97.
    --- mogwaï (R: 0.46; S: 0.98)
    --- gizmo (R: 0.44; S: 0.96)
    --- is (R: 0.13; S: 0.98)
    -- F: 1.00. R: 0.18. S: 0.21.
    --- in (R: 0.05; S: 0.21)
    --- chinese (R: 0.05; S: 0.21)
    --- folklore (R: 0.05; S: 0.21)
    --- demon (R: 0.05; S: 0.21)
    - F: 0.60. R: 0.04. S: 0.12.
    -- blade (R: 0.01; S: 0.03)
    -- F: 1.00. R: 0.02. S: 0.15.
    --- sentence (R: 0.01; S: 0.15)
    --- about (R: 0.01; S: 0.15)
    """

    def __init__(self, corpus, embedding):
        self.corpus = corpus
        self.embedding = embedding
        self.diteration = DIteration(n=embedding.n, m=embedding.m)

        self.post_document = post_document
        self.post_feature = post_feature
        self.post_document_cluster = post_document_cluster
        self.post_feature_cluster = post_feature_cluster

    # Ranking Part
    def rank(self, query=""):
        """
        Runs the Diteration using query as starting point

        Parameters
        ----------
        query: str
               Text that starts DIteration
        """
        z = self.embedding.query_projection(query)
        self.diteration(self.embedding.x, self.embedding.y, z)

    def get_ranked_documents(self, k=10):
        return [self.post_document(self, i) for i in self.diteration.x_order[:k]]

    def get_ranked_features(self, k=10):
        return [self.post_feature(self, i) for i in self.diteration.y_order[:k]]

    # Cluster part
    def get_clustered_documents(self, indices, resolution=.9):
        subspace = csr_matrix(vstack([self.embedding.x[i, :].multiply(self.diteration.y_relevance) for i in indices]))
        cluster = subspace_clusterize(subspace, resolution, indices)
        return self.post_document_cluster(self, cluster)

    def get_clustered_ranked_documents(self, k=10, resolution=.9):
        return self.get_clustered_documents(self.diteration.x_order[:k], resolution)

    def get_clustered_features(self, indices, resolution=.9):
        subspace = csr_matrix(vstack([self.embedding.y[i, :].multiply(self.diteration.x_relevance) for i in indices]))
        cluster = subspace_clusterize(subspace, resolution, indices)
        return self.post_feature_cluster(self, cluster)

    def get_clustered_ranked_features(self, k=10, resolution=.9):
        return self.get_clustered_features(self.diteration.y_order[:k], resolution)

